from typing import Dict, Any

import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql


@commands.load.register(Context, Sql, dict)
def load(context: Context, backend: Sql, config: Dict[str, Any]):
    dsn = config['dsn']
    schema = config.get('schema')
    if dsn:
        # There can be a situation, when backend type was given, but dsn was
        # not. In this case we still create backend instance, but leave it
        # uninitialized.
        backend.engine = sa.create_engine(dsn, echo=False)
        backend.schema = sa.MetaData(backend.engine, schema=schema)
        backend.dbschema = schema



from typing import List
from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Mode
from spinta.core.config import KeyFormat
from spinta.core.context import configure_context


def config(
    ctx: TyperContext,
    name: Optional[List[str]] = Argument(None),
    fmt: KeyFormat = KeyFormat.cfg,
):
    """Show current configuration values"""
    context = configure_context(ctx.obj)
    rc = context.get('rc')
    rc.dump(*name, fmt=fmt)


def check(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
    mode: Mode = Option('internal', help="Mode of backend operation"),
):
    """Check configuration and manifests"""
    context = configure_context(ctx.obj, manifests, mode=mode)
    prepare_manifest(context, ensure_config_dir=True)
    echo("OK")



from spinta import commands
from spinta.components import Context, Model
from spinta.backends.mongo.components import Mongo
from spinta.types.datatype import DataType


@commands.wipe.register(Context, Model, Mongo)
def wipe(context: Context, model: Model, backend: Mongo):
    for prop in model.properties.values():
        wipe(context, prop.dtype, backend)

    table_main = backend.db[model.model_type()]
    table_changelog = backend.db[model.model_type() + '__changelog']
    table_main.delete_many({})
    table_changelog.delete_many({})


@commands.wipe.register(Context, DataType, Mongo)
def wipe(context: Context, dtype: DataType, backend: Mongo):
    same_backend = backend.name == dtype.backend.name

    if not same_backend:
        wipe(context, dtype, dtype.backend)



import inspect
import importlib
import pathlib
import time
import types

import itertools

from spinta import commands
from spinta.backends.components import BackendOrigin
from spinta.backends.helpers import load_backend
from spinta.components import Context, Store
from spinta.urlparams import get_model_by_name
from spinta.manifests.helpers import create_manifest
from spinta.manifests.helpers import create_internal_manifest


@commands.load.register(Context, Store)
def load(context: Context, store: Store) -> Store:
    """Load backends and manifests from configuration."""

    rc = context.get('rc')
    config = context.get('config')

    # Load keymaps
    store.keymaps = {}
    for name in rc.keys('keymaps'):
        keymap_type = rc.get('keymaps', name, 'type', required=True)
        KeyMap = config.components['keymaps'][keymap_type]
        keymap = store.keymaps[name] = KeyMap()
        keymap.name = name
        commands.configure(context, keymap)

    # Load backends
    # Backends can also be loaded from manifests, see:
    # `spinta.manifests.helpers._load_manifest_backends`.
    store.backends = {}
    for name in rc.keys('backends'):
        data = rc.to_dict('backends', name)
        store.backends[name] = load_backend(
            context,
            config,
            name,
            BackendOrigin.config,
            data,
        )

    # Create default manifest instance
    manifest = rc.get('manifest', required=True)
    manifest = create_manifest(context, store, manifest)
    store.manifest = manifest

    # Create internal manifest instance
    store.internal = create_internal_manifest(context, store)

    # Load accesslog
    store.accesslog = commands.load(context, config.AccessLog(), config)

    return store


@commands.wait.register(Context, Store)
def wait(
    context: Context,
    store: Store,
    *,
    seconds: int = None,
    verbose: bool = False,
) -> bool:
    rc = context.get('rc')

    if seconds is None:
        seconds = rc.get('wait', cast=int, required=True)

    # Collect all backends
    backends = set(itertools.chain(
        store.backends.values(),
        store.manifest.backends.values(),
        (
            resource.backend
            for dataset in store.manifest.datasets.values()
            for resource in dataset.resources.values()
            if resource.backend
        )
    ))

    # XXX: Probably whole this has to be moved to cli
    # Wait while all backends are up.
    i = 0
    fail = False
    start = time.time()
    while backends:
        i += 1
        fail = time.time() - start > seconds
        if verbose:
            print(f"Waiting for backends, attempt #{i}:")
        for backend in sorted(backends, key=lambda b: b.name):
            if verbose:
                print(f"  {backend.name}...")
            if commands.wait(context, backend, fail=fail):
                backends.remove(backend)
        if fail:
            break
        else:
            time.sleep(1)

    if verbose:
        if fail:
            print("Timeout. Not all backends are ready in given time.")
        else:
            print("All backends are up.")

    return True


@commands.push.register(Context, Store, types.GeneratorType)
def push(context: Context, store: Store, stream: types.GeneratorType):
    manifest = store.manifest
    for data in stream:
        data = dict(data)
        model_name = data.pop('type', None)
        assert model_name is not None, data
        model = get_model_by_name(context, manifest, model_name)
        if 'id' in data:
            id_ = commands.upsert(context, model, model.backend, key=['id'], data=data)
        else:
            id_ = commands.insert(context, model, model.backend, data=data)
        if id_ is not None:
            yield id_


@commands.push.register(Context, Store, list)
def push(context: Context, store: Store, stream: list):
    yield from push(context, store, (x for x in stream))


def find_subclasses(Class, modules):
    for module_path in modules:
        module = importlib.import_module(module_path)
        path = pathlib.Path(module.__file__).parent
        base = path.parents[module_path.count('.')]
        for path in path.glob('**/*.py'):
            if path.name == '__init__.py':
                module_path = path.parent.relative_to(base)
            else:
                module_path = path.relative_to(base).with_suffix('')
            module_path = '.'.join(module_path.parts)
            module = importlib.import_module(module_path)
            for obj_name in dir(module):
                obj = getattr(module, obj_name)
                if inspect.isclass(obj) and issubclass(obj, Class) and obj is not Class and obj.__module__ == module_path:
                    yield obj



from enum import Enum
from typing import Any
from typing import Optional
from typing import Type

from spinta import exceptions
from spinta.components import Component


def get_enum_by_name(enum, value):
    for item in enum:
        if item.name == value:
            return item
    raise Exception(f"Unknown value {value!r}.")


def enum_by_name(
    component: Component,
    param: Optional[str],  # component param
    enum: Type[Enum],
    name: Any,
) -> Optional[Enum]:
    if name is None or name == '':
        return None
    for item in enum:
        if item.name == name:
            return item
    raise exceptions.InvalidValue(component, param=param, given=name)



from typing import Dict

from spinta import commands
from spinta.datasets.components import Resource, Entity, Attribute


@commands.get_error_context.register(Resource)
def get_error_context(resource: Resource, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(resource.dataset, prefix=f'{prefix}.dataset')
    context['resource'] = f'{prefix}.name'
    return context


@commands.get_error_context.register(Entity)
def get_error_context(external: Entity, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(external.model, prefix=f'{prefix}.model')
    context['external'] = f'{prefix}.name'
    return context


@commands.get_error_context.register(Attribute)
def get_error_context(external: Attribute, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(external.prop, prefix=f'{prefix}.prop')
    context['external'] = f'{prefix}.name'
    return context



from tabulate import tabulate


def draw(buffer, name, tnum, width):
    if tnum > 1:
        if name:
            yield f"\n\nTable {name} #{tnum}:\n"
        else:
            yield f"\n\nTable #{tnum}:\n"

    yield tabulate(buffer, missingval='∅')



from spinta import commands
from spinta.components import Context
from spinta.manifests.tabular.components import TabularManifest


@commands.bootstrap.register(Context, TabularManifest)
def bootstrap(context: Context, manifest: TabularManifest):
    # Tabular manifest can't store state so we always run bootstrap.
    store = context.get('store')
    for backend in store.backends.values():
        commands.bootstrap(context, backend)



try:
    from alembic.migration import MigrationContext
    from alembic.operations import Operations
except ImportError:
    pass

from spinta import commands
from spinta.core.ufuncs import asttoexpr, ufunc
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.backends.postgresql.components import PostgreSQL


@commands.migrate.register(Context, Manifest, PostgreSQL)
def migrate(context: Context, manifest: Manifest, backend: PostgreSQL):
    config = context.get('config')
    resolvers = ufunc.resolver.ufuncs()
    executors = ufunc.executor.ufuncs()

    Env = config.components['migrations']
    conn = context.get(f'transaction.{backend.name}')
    ctx = MigrationContext.configure(conn)
    op = Operations(ctx)
    scope = {
        'op': op,
        'backend': backend
    }

    async def execute(versions):
        for version in versions:
            for action in version['actions']:
                ast = action['upgrade']
                env = Env(context, resolvers, executors, scope)
                expr = asttoexpr(ast)
                expr = env.resolve(expr)
                env.execute(expr)
            yield version

    return execute



import sqlalchemy as sa
import sqlalchemy.exc

from spinta import commands
from spinta.components import Context
from spinta.backends.postgresql.components import PostgreSQL


@commands.wait.register(Context, PostgreSQL)
def wait(context: Context, backend: PostgreSQL, *, fail: bool = False) -> bool:
    rc = context.get('rc')
    dsn = rc.get('backends', backend.name, 'dsn', required=True)
    engine = sa.create_engine(dsn, connect_args={'connect_timeout': 0})
    try:
        conn = engine.connect()
    except sqlalchemy.exc.OperationalError:
        if fail:
            raise
        else:
            return False
    else:
        conn.close()
        engine.dispose()
        return True



import pathlib
from typing import Optional

from spinta.auth import gen_auth_server_keys
from spinta.core.config import RawConfig
from spinta.testing.datasets import Sqlite
from spinta.testing.tabular import create_tabular_manifest
from spinta.manifests.tabular.helpers import striptable

CONFIG = {
    'environments': {
        'test': {
            'backends': {
                'default': {
                    'type': 'postgresql',
                    'dsn': 'postgresql://admin:admin123@localhost:54321/spinta_tests',
                },
                'mongo': {
                    'type': 'mongo',
                    'dsn': 'mongodb://admin:admin123@localhost:27017/',
                    'db': 'spinta_tests',
                },
                'fs': {
                    'type': 'fs',
                    'path': pathlib.Path() / 'var/files',
                },
            },
        },
    },
}


def create_config_path(path: pathlib.Path) -> pathlib.Path:
    path.mkdir(exist_ok=True)
    (path / 'clients').mkdir(exist_ok=True)
    gen_auth_server_keys(path, exist_ok=True)
    return path


def configure(
    rc: RawConfig,
    db: Optional[Sqlite],   # TODO: move backend configuration to manifest
    path: pathlib.Path,     # manifest file path
    manifest: str,          # manifest as ascii table string
) -> RawConfig:
    create_tabular_manifest(path, striptable(manifest))
    if db:
        return rc.fork({
            'manifests': {
                'default': {
                    'type': 'tabular',
                    'path': str(path),
                    'backend': 'sql',
                    'keymap': 'default',
                    'mode': 'external',
                },
            },
            'backends': {
                'sql': {
                    'type': 'sql',
                    'dsn': db.dsn,
                },
            },
            # tests/config/clients/3388ea36-4a4f-4821-900a-b574c8829d52.yml
            'default_auth_client': '3388ea36-4a4f-4821-900a-b574c8829d52',
        })
    else:
        return rc.fork({
            'manifests': {
                'default': {
                    'type': 'tabular',
                    'path': str(path),
                    'backend': 'default',
                    'keymap': 'default',
                    'mode': 'external',
                },
            },
            'backends': {
                'default': {
                    'type': 'memory',
                },
            },
            # tests/config/clients/3388ea36-4a4f-4821-900a-b574c8829d52.yml
            'default_auth_client': '3388ea36-4a4f-4821-900a-b574c8829d52',
        })



from spinta.commands import check, load
from spinta.components import Context, MetaData
from spinta.manifests.components import Manifest
from spinta.nodes import load_node
from spinta import exceptions


class Owner(MetaData):
    schema = {
        'title': {'type': 'string'},
        'sector': {'type': 'string'},
        'logo': {'type': 'path'},
        'backend': {'type': 'backend', 'inherit': True, 'required': False},
    }

    def __init__(self):
        super().__init__()
        self.logo = None
        self.sector = None


@load.register(Context, Owner, dict, Manifest)
def load(
    context: Context,
    owner: Owner,
    data: dict,
    manifest: Manifest,
    *,
    source: Manifest = None,
):
    return load_node(context, owner, data)


@check.register(Context, Owner)
def check(context: Context, owner: Owner):
    if owner.logo:
        path = owner.manifest.path / 'media/owners' / owner.name / owner.logo
        if not path.exists():
            raise exceptions.FileNotFound(str(path), owner)



from typing import Generic
from typing import TypeVar, AsyncIterator, List, Callable, Optional, Tuple, Awaitable, Iterable

T = TypeVar('T')    # iterator item type
K = TypeVar('K')    # type of a group by item


class agroupby(Generic[T, K]):
    # Adapted to async from Python docs:
    # https://docs.python.org/3/library/itertools.html#itertools.groupby

    keyfunc: Callable[[T], K]
    it: AsyncIterator[T]
    tgtkey: K
    currkey: K
    currvalue: T

    def __init__(
        self,
        iterable: AsyncIterator[T],
        key: Optional[Callable[[T], K]] = None,
    ) -> None:
        if key is None:
            key = lambda x: x  # noqa
        self.keyfunc = key
        self.it = iterable
        self.tgtkey = self.currkey = self.currvalue = object()

    def __aiter__(self) -> AsyncIterator[Tuple[K, AsyncIterator[T]]]:
        return self

    async def __anext__(self) -> Awaitable[Tuple[K, AsyncIterator[T]]]:
        self.id = object()
        while self.currkey == self.tgtkey:
            self.currvalue = await self.it.__anext__()    # Exit on AsyncStopIteration
            self.currkey = self.keyfunc(self.currvalue)
        self.tgtkey = self.currkey
        return self.currkey, self._grouper(self.tgtkey, self.id)

    async def _grouper(self, tgtkey: K, id_: object) -> AsyncIterator[T]:
        while self.id is id_ and self.currkey == tgtkey:
            yield self.currvalue
            try:
                self.currvalue = await self.it.__anext__()
            except StopAsyncIteration:
                return
            self.currkey = self.keyfunc(self.currvalue)


async def alist(it: AsyncIterator[T]) -> List[T]:
    return [x async for x in it]


async def aiter(it: Iterable[T]) -> AsyncIterator[T]:
    for x in it:
        yield x


async def aslice(it: AsyncIterator[T], *args: int) -> AsyncIterator[T]:
    i = 0
    r = range(*args)
    async for x in it:
        if i in r:
            yield x
        i += 1


async def adrain(it: AsyncIterator[T]) -> None:
    async for _ in it:
        pass



from typing import Dict
from typing import Union

import contextlib
import itertools
import uuid

import sqlalchemy as sa
from sqlalchemy.engine import Engine

from spinta.backends.postgresql.helpers import get_column_name
from spinta.types.datatype import Ref
from spinta.utils.schema import NA
from spinta.components import Model, Property
from spinta.backends.constants import TableType
from spinta.backends.components import Backend, BackendFeatures
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.sqlalchemy import utcnow
from spinta.exceptions import MultipleRowsFound, NotFoundError
from spinta.exceptions import PropertyNotFound


class PostgreSQL(Backend):
    metadata = {
        'name': 'postgresql',
        'properties': {
            'dsn': {'type': 'string', 'required': True},
        },
    }

    features = {
        BackendFeatures.FILE_BLOCKS,
        BackendFeatures.WRITE,
    }

    engine: Engine = None
    schema: sa.MetaData = None
    tables: Dict[str, sa.Table] = None

    @contextlib.contextmanager
    def transaction(self, write=False):
        with self.engine.begin() as connection:
            if write:
                table = self.tables['_txn']
                result = connection.execute(
                    table.insert().values(
                        # FIXME: commands.gen_object_id should be used here
                        _id=str(uuid.uuid4()),
                        datetime=utcnow(),
                        client_type='',
                        client_id='',
                        errors=0,
                    )
                )
                transaction_id = result.inserted_primary_key[0]
                yield WriteTransaction(connection, transaction_id)
            else:
                yield ReadTransaction(connection)

    @contextlib.contextmanager
    def begin(self):
        with self.engine.begin() as conn:
            yield conn

    def get(self, connection, columns, condition, default=NA):
        scalar = isinstance(columns, sa.Column)
        columns = columns if isinstance(columns, list) else [columns]

        result = connection.execute(
            sa.select(columns).where(condition)
        )
        result = list(itertools.islice(result, 2))

        if len(result) == 1:
            if scalar:
                return result[0][columns[0]]
            else:
                return result[0]

        elif len(result) == 0:
            if default is NA:
                raise NotFoundError()
            else:
                return default
        else:
            raise MultipleRowsFound()

    def add_table(
        self,
        table: sa.Table,
        node: Union[Model, Property],
        ttype: TableType = TableType.MAIN,
    ):
        name = get_table_name(node, ttype)
        assert name not in self.tables, name
        self.tables[name] = table

    def get_table(
        self,
        node: Union[Model, Property],
        ttype: TableType = TableType.MAIN,
        *,
        fail: bool = True,
    ):
        name = get_table_name(node, ttype)
        if fail:
            return self.tables[name]
        else:
            return self.tables.get(name)

    def get_column(self, table: sa.Table, prop: Property, *, select=False) -> sa.Column:
        if prop.list is not None:
            table = self.get_table(prop.list, TableType.LIST)
        column = get_column_name(prop)
        if isinstance(prop.dtype, Ref):
            # TODO: Move this into a command.
            column += '._id'
        if column not in table.c:
            raise PropertyNotFound(prop.dtype)
        return table.c[column]

    def query_nodes(self):
        return []

    def bootstrapped(self):
        meta = sa.MetaData(self.engine)
        table = sa.Table('_schema', meta)
        insp = sa.inspect(self.engine)
        if insp.has_table(table.name):
            with self.engine.begin() as conn:
                query = sa.select([sa.func.count()]).select_from(table)
                return conn.execute(query).scalar() > 0
        return False


class ReadTransaction:
    id: str
    errors: int

    def __init__(self, connection):
        self.connection = connection


class WriteTransaction(ReadTransaction):

    def __init__(self, connection, id_: str):
        super().__init__(connection)
        self.id = id_
        self.errors = 0



from typing import List
from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Mode
from spinta.core.context import configure_context
from spinta.manifests.tabular.helpers import render_tabular_manifest


def show(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
    mode: Mode = Option('internal', help="Mode of backend operation"),
):
    """Show manifest as ascii table"""
    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context, verbose=False)
    manifest = store.manifest
    echo(render_tabular_manifest(manifest))




from __future__ import annotations

import operator
from textwrap import indent
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Union
from typing import cast
from typing import NamedTuple
from typing import Optional

import httpx
import requests
from pprintpp import pformat

from spinta.formats.html.components import Cell
from spinta.testing.client import TestClient
from spinta.utils.data import take
from spinta.utils.nestedstruct import flatten


def listdata(
    resp: Union[requests.Response, List[Dict[str, Any]]],
    *keys: Union[str, Callable[[], bool]],
    sort: Union[bool, str] = True,
    full: bool = False,  # returns dicts instead of tuples
) -> List[tuple]:
    """Return data from a given requests.Response object.

    Only non reserved fields are returned.

    By default data are converted to List[Tuple[Any]], but if `full` is True,
    then List[Dict[str, Any] is returned.

    Usage:

        >>> data = [
        ...     {'_id': 1, 'odd': 1, 'even': 2},
        ...     {'_id': 2, 'odd': 3, 'even': 4},
        ...     {'_id': 3, 'odd': 5, 'even': 6},
        ... }

        >>> listdata(data)
        [
            (1, 2),
            (3, 4),
            (5, 6),
        ]

        >>> listdata(data, 'even')
        [2, 4, 6]

        >>> listdata(data, 'odd', 'even')
        [
            (1, 2),
            (3, 4),
            (5, 6),
        ]

        >>> listdata(data, full=True)
        data = [
            {'odd': 1, 'even': 2},
            {'odd': 3, 'even': 4},
            {'odd': 5, 'even': 6},
        }

    """

    # Prepare data
    if isinstance(resp, list):
        data = resp

    elif resp.headers['content-type'].startswith('text/html'):
        data = resp.context
        assert resp.status_code == 200, pformat(data)
        assert 'data' in data, pformat(data)
        assert 'header' in data, pformat(data)
        header = data['header']
        keys = keys or [k for k in header if not k.startswith('_')]
        data = [
            {k: v.value for k, v in zip(header, row)}
            for row in cast(List[List[Cell]], data['data'])
        ]

    else:
        data = resp.json()
        assert resp.status_code == 200, pformat(data)
        assert '_data' in data, pformat(data)
        data = data['_data']
        keys = keys or sorted({
            k
            for d in flatten(data)
            for k in d
            if not k.startswith('_')
        })

    # Clean data
    if full:
        data = [take(keys, row) for row in data]
    elif len(keys) == 1:
        k = keys[0]
        data = [take(k, row) for row in data]
    else:
        data = [tuple(take(k, row) for k in keys) for row in data]

    # Sort
    if sort is True:
        data = sorted(data, key=str)
    elif sort:
        if full:
            sort_key = operator.itemgetter(sort)
        else:
            sort_key = operator.itemgetter(keys.index(sort))
        data = sorted(data, key=sort_key)

    return data


def pushdata(app: TestClient, *args) -> Union[dict, list]:
    model: str
    data: Union[list, dict]
    if len(args) == 2:
        model, data = args
    else:
        model = '/'
        data, = args
    multiple = isinstance(data, list)
    if multiple:
        data = {'_data': data}
        resp = app.post(model, json=data)
    else:
        resp = app.post(model, json=data)
    try:
        resp.raise_for_status()
    except requests.HTTPError as e:
        dump = indent(pformat(data), '  ').strip()
        raise Exception(
            f"pushdata error:\n"
            f"  model={model},\n"
            f"  error={e},\n  "
            f"data: {dump}"
        )
    data = resp.json()
    if multiple:
        return sorted(data['_data'], key=lambda x: x['_id'])
    else:
        return data


class Data(NamedTuple):
    id: str
    rev: str
    data: Dict[str, Any]

    @property
    def sid(self):
        if self.id:
            return self.id[:8]

    def update(self, data: Dict[str, Any]) -> Data:
        return Data(id=self.id, rev=self.rev, data={**self.data, **data})


def _obj_from_dict(data: Dict[str, Any]) -> Data:
    return Data(
        id=data.pop('_id', None),
        rev=data.pop('_revision', None),
        data=data,
    )


def send(
    app: TestClient,
    model: str,
    action: str,
    obj: Optional[Union[Data, Dict[str, Any]]] = None,
    data: Optional[Dict[str, Any]] = None,
    *,
    select: Optional[List[str]] = None,
    sort: Optional[Union[bool, str]] = None,
    full: Optional[bool] = None,  # returns dicts instead of tuples
) -> Union[Data, List[Data]]:
    if obj is not None and isinstance(obj, dict):
        obj = _obj_from_dict(obj)

    if data is not None:
        obj = obj.update(data)

    if action == 'insert':
        resp = app.post(model, json=obj.data)
    elif action in 'patch':
        resp = app.patch(f'{model}/{obj.id}', json={
            **obj.data,
            '_revision': obj.rev,
        })
    elif action in 'delete':
        resp = app.delete(f'{model}/{obj.id}')
    elif action == 'changes':
        if obj:
            resp = app.get(f'{model}/{obj.id}/:changes')
        else:
            resp = app.get(f'{model}/:changes')
    elif action == 'getall':
        resp = app.get(model)
    elif action == 'getone':
        resp = app.get(f'{model}/{obj.id}')
    else:
        resp = app.get(f'{model}/{action}')

    if select is not None or sort is not None or full is not None:
        if select is None:
            select = []
        if sort is None:
            sort = False
        if full is None:
            full = True
        return listdata(resp, *select, sort=sort, full=full)

    try:
        resp.raise_for_status()
    except httpx.HTTPStatusError as e:
        dump = indent(pformat(obj.data), '  ').strip()
        raise Exception(
            f"send error:\n"
            f"  model={model},\n"
            f"  action={action},\n"
            f"  error={e},\n  "
            f"data: {dump}"
        )

    result = resp.json()

    if '_data' in result:
        return [_obj_from_dict(data) for data in result['_data']]

    else:
        return _obj_from_dict(result)




from spinta import commands
from spinta.types.datatype import Object
from spinta.components import Context, DataSubItem, Action
from spinta.backends.mongo.components import Mongo
from spinta.utils.schema import NA


@commands.before_write.register(Context, Object, Mongo)
def before_write(
    context: Context,
    dtype: Object,
    backend: Mongo,
    *,
    data: DataSubItem,
) -> dict:
    patch = {}
    if data.root.action == Action.INSERT or (data.root.action == Action.UPSERT and data.saved is NA):
        return {dtype.prop.place: data.patch}

    for prop in dtype.properties.values():
        value = commands.before_write(
            context,
            prop.dtype,
            backend,
            data=data[prop.name],
        )
        patch.update(value)
    return patch



import contextlib
from pathlib import Path
from typing import TextIO

from spinta.datasets.components import ExternalBackend


class SqlDump(ExternalBackend):
    """
    If SQL dump is given via stdin, then path is None and stream is assigned to
    stdin. Otherwise stream is None and path should be defined.
    """

    type: str = 'sqldump'
    path: Path = None
    stream: TextIO = None

    @contextlib.contextmanager
    def transaction(self, write=False):
        raise NotImplementedError

    @contextlib.contextmanager
    def begin(self):
        raise NotImplementedError




import re
from typing import Set

from unidecode import unidecode

camel_cased_words = re.compile(r'([A-Z][a-z]+)')
split_words_re = re.compile(r'[^a-zA-Z0-9]+')
number_prefix_re = re.compile(r'^([0-9]+)')


def _cleanup(name: str) -> str:
    name = number_prefix_re.sub(r'N\1', name)
    return name


def to_model_name(name: str) -> str:
    name = unidecode(name)
    name = camel_cased_words.sub(r' \1', name).strip()
    words = split_words_re.split(name)
    if name.isupper() or name.islower():
        words = [w.title() for w in words if w]
    else:
        words = [w.title() if w.islower() else w for w in words if w]
    return _cleanup(''.join(words))


def to_property_name(name: str, is_ref: bool = False) -> str:
    name = to_code_name(name)
    if is_ref and name.endswith('_id'):
        name = name[:-3]
    return name


def to_code_name(name: str) -> str:
    name = unidecode(name)
    name = camel_cased_words.sub(r' \1', name).strip()
    words = split_words_re.split(name)
    words = filter(None, words)
    return _cleanup('_'.join(words)).lower()


class Deduplicator:
    _names: Set[str]

    def __init__(self, template: str = '{}'):
        self._names = set()
        self._template = template

    def __call__(self, name: str):
        name_ = name
        i = 0
        while name_ in self._names:
            i += 1
            name_ = name + self._template.format(i)
        self._names.add(name_)
        return name_



from typing import List

import cgi
from typing import Union

import itertools
import urllib.parse

from starlette.requests import Request

from spinta.commands import prepare
from spinta.components import Action
from spinta.components import Config
from spinta.components import Context, Node
from spinta.components import Model
from spinta.components import Namespace
from spinta.manifests.components import Manifest
from spinta.utils import url as urlutil
from spinta.components import UrlParams, Version
from spinta.commands import is_object_id
from spinta import exceptions
from spinta import spyna
from spinta.exceptions import ModelNotFound


@prepare.register(Context, UrlParams, Version, Request)
def prepare(
    context: Context,
    params: UrlParams,
    version: Version,
    request: Request,
) -> UrlParams:
    params.parsetree = (
        urlutil.parse_url_path(request.path_params['path'].strip('/')) +
        parse_url_query(urllib.parse.unquote(request.url.query))
    )
    prepare_urlparams(context, params, request)
    params.head = request.method == 'HEAD'
    params.action = get_action(params, request)
    return params


METHOD_TO_ACTION = {
    'POST': Action.INSERT,
    'PUT': Action.UPDATE,
    'PATCH': Action.PATCH,
    'DELETE': Action.DELETE,
}


def get_action(params: UrlParams, request: Request) -> Action:
    if params.action is not None:
        # Might be set in _prepare_urlparams_from_path
        return params.action
    if request.method in ('GET', 'HEAD'):
        if params.changes:
            return Action.CHANGES
        elif params.pk:
            return Action.GETONE
        else:
            search = any((
                params.query,
                params.select,
                params.limit is not None,
                params.offset is not None,
                params.sort,
                params.count,
            ))
            return Action.SEARCH if search else Action.GETALL
    else:
        return METHOD_TO_ACTION[request.method]


def parse_url_query(query):
    if not query:
        return []
    rql = spyna.parse(query)
    if rql['name'] == 'and':
        return rql['args']
    else:
        return [rql]


def prepare_urlparams(context: Context, params: UrlParams, request: Request):
    _prepare_urlparams_from_path(params)
    _resolve_path(context, params)
    params.format = get_response_type(context, request, params)
    config: Config = context.get('config')
    params.fmt = config.exporters[params.format]


def _prepare_urlparams_from_path(params: UrlParams):
    params.formatparams = {}

    for param in params.parsetree:
        name = param['name']
        args = param['args']

        if name == 'path':
            params.path_parts = args
        elif name == 'ns':
            params.ns = True
        elif name == 'all':
            params.all = True
        elif name == 'select':
            # TODO: Traverse args and resolve all properties, etc.
            # XXX: select can be used only once.
            if params.select is None:
                params.select = args
            else:
                params.select += args
            # TODO: This is a temporary hack, there is no need to specifically
            #       mark count in params.
            for arg in args:
                if arg['name'] == 'count':
                    params.count = True
                    break
        elif name == 'sort':
            # TODO: Traverse args and resolve all properties, etc.
            if params.sort is None:
                params.sort = args
            else:
                params.sort += args
        elif name == 'limit':
            if params.limit is not None:
                raise exceptions.InvalidValue(
                    operator='limit',
                    message="Limit operator can be set only once.",
                )
            if len(args) == 2:
                if params.offset is not None:
                    raise exceptions.InvalidValue(
                        operator='offset',
                        message="Offset operator can be set only once.",
                    )
                params.limit, params.offset = map(int, args)
            elif len(args) == 1:
                params.limit = int(args[0])
            else:
                raise exceptions.InvalidValue(
                    operator='limit',
                    message="Too many arguments.",
                )
        elif name == 'offset':
            if params.offset is not None:
                raise exceptions.InvalidValue(
                    operator='offset',
                    message="Offset operator can be set only once.",
                )
            if len(args) == 1:
                params.offset = int(args[0])
            else:
                raise exceptions.InvalidValue(
                    operator='offset',
                    message="Too many or too few arguments. One argument is expected.",
                )
        # XXX: `count` can't be at the top level, well it can, but at the top
        #      level it is used as where condition.
        elif name == 'count':
            params.count = True
        elif name == 'changes':
            params.changes = True
            params.changes_offset = int(args[0]) if args else None
        elif name == 'format':
            if isinstance(args[0], dict) and args[0]['name'] == 'bind':
                params.format = _read_format_params(args[0])
                args = args[1:]
            elif isinstance(args[0], str):
                params.format = args[0]
                args = args[1:]
            for arg in args:
                assert len(arg['args']) == 1, arg
                params.formatparams[arg['name']] = _read_format_params(arg['args'][0])
        elif name == 'summary':
            params.summary = True
        elif name == 'fault-tolerant':
            params.fault_tolerant = True
        elif name == 'wipe':
            params.action = Action.WIPE
        elif name == 'check':
            params.action = Action.CHECK
        else:
            if params.query is None:
                params.query = []
            params.query.append(param)


def _read_format_params(node):
    if not isinstance(node, dict):
        return node

    if node['name'] == 'bind':
        return node['args'][0]

    raise Exception(f"Unknown node {node!r}.")


def _join_path_parts(*parts: str) -> str:
    return '/'.join(parts)


def _find_model_name_index(
    manifest: Manifest,
    parts: List[str],
) -> int:
    keys = (
        set(manifest.namespaces) |
        set(manifest.models) |
        set(manifest.endpoints)
    )
    for i, name in enumerate(itertools.accumulate(parts, _join_path_parts)):
        if name not in keys:
            return i
    return len(parts)


def _resolve_path(context: Context, params: UrlParams) -> None:
    if params.path_parts is None:
        params.path = ''
        params.path_parts = []

    manifest = context.get('store').manifest
    i = _find_model_name_index(manifest, params.path_parts)
    parts = params.path_parts[i:]
    params.path = '/'.join(params.path_parts[:i])
    params.model = get_model_from_params(manifest, params)

    if parts:
        # Resolve ID.
        params.pk = parts.pop(0)
        if not is_object_id(context, params.model.backend, params.model, params.pk):
            given_path = '/'.join(params.path_parts)
            raise ModelNotFound(model=given_path)

    if parts:
        # Resolve property (subresource).
        prop = parts.pop(0)
        if prop.endswith(':ref'):
            params.propref = True
            prop = prop[:-len(':ref')]
        if prop not in params.model.flatprops:
            raise exceptions.PropertyNotFound(params.model, property=prop)
        params.prop = params.model.flatprops[prop]

    if parts:
        given_path = '/'.join(params.path_parts)
        raise ModelNotFound(model=given_path)


def get_model_by_name(context: Context, manifest: Manifest, name: str) -> Node:
    config = context.get('config')
    UrlParams_ = config.components['urlparams']['component']
    params = UrlParams_()
    params.parsetree = urlutil.parse_url_path(name)
    _prepare_urlparams_from_path(params)
    _resolve_path(context, params)
    return params.model


def get_model_from_params(
    manifest: Manifest,
    params: UrlParams,
) -> Union[Namespace, Model]:
    name = params.path

    if name in manifest.endpoints:
        # Allow users to specify a different URL endpoint to make URL look
        # nicer, but that is optional, they can still use model.name.
        name = manifest.endpoints[name]

    if params.ns:
        if name in manifest.namespaces:
            return manifest.namespaces[name]
        else:
            raise ModelNotFound(manifest, model=name)

    elif name in manifest.models:
        return manifest.models[name]

    elif name in manifest.namespaces:
        return manifest.namespaces[name]

    else:
        raise ModelNotFound(model=name)


def get_response_type(
    context: Context,
    request: Request,
    params: UrlParams = None,
) -> str:
    config = context.get('config')

    if params is not None and params.format:
        if params.format not in config.exporters:
            raise exceptions.UnknownOutputFormat(name=params.format)
        return params.format

    if 'accept' in request.headers and request.headers['accept']:
        formats = {
            'text/html': 'html',
            'application/xhtml+xml': 'html',
        }
        for name, exporter in config.exporters.items():
            for media_type in exporter.accept_types:
                formats[media_type] = name

        media_types, _ = cgi.parse_header(request.headers['accept'])
        for media_type in media_types.lower().split(','):
            if media_type in formats:
                return formats[media_type]

    return 'json'



import pathlib
import os

from spinta import commands
from spinta.components import Context, Action, Property, DataItem
from spinta.backends.components import Backend
from spinta.types.datatype import DataType, File
from spinta.backends.fs.components import FileSystem
from spinta.exceptions import FileNotFound, ConflictingValue, UnacceptableFileName


@commands.simple_data_check.register(Context, DataItem, File, Property, FileSystem, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: File,
    prop: Property,
    backend: FileSystem,
    value: dict,
):
    if value['_id'] is not None:
        # Check if given filepath stays on backend.path.
        _validate_path(value['_id'], backend, dtype)


@commands.complex_data_check.register()
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: File,
    prop: Property,
    backend: Backend,
    given: dict,
):
    complex_data_check[
        type(context),
        DataItem,
        DataType,
        Property,
        Backend,
        dict,
    ](context, data, dtype, prop, backend, given)
    if isinstance(dtype.backend, FileSystem):
        _validate_path(given['_id'], dtype.backend, dtype)
        path = dtype.backend.path / given['_id']
        if '_content' not in given and not path.exists():
            raise FileNotFound(prop, file=given['_id'])


@commands.complex_data_check.register()
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: File,
    prop: Property,
    backend: FileSystem,
    value: dict,
):
    # TODO: revision check for files
    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    dtype.prop,
                    given=data.given[k],
                    expected=data.saved[k],
                )
        if value.get('_id'):
            if isinstance(dtype.backend, FileSystem):
                filename = pathlib.PosixPath(value['_id'])
                _validate_path(filename, dtype.backend, dtype)


def _validate_path(filename: pathlib.PosixPath(), fs: FileSystem, dtype: File):
    commonpath = os.path.commonpath([
        fs.path.resolve(),
        (fs.path / filename).resolve(),
    ])
    if str(commonpath) != str(fs.path.resolve()):
        raise UnacceptableFileName(dtype, file=filename)
